@@ -1,10 +1,11 @@
 import re
-from typing import Any
+from typing import Any, NoReturn
 
 import pytest
 from yarl import URL
 
-from aiohttp import web
+from aiohttp import web, web_app
+from aiohttp.pytest_plugin import AiohttpClient
 from aiohttp.typedefs import Handler
 
 
@@ -24,10 +25,13 @@ async def test_middleware_modifies_response(loop, aiohttp_client) -> None:
     app.middlewares.append(middleware)
     app.router.add_route("GET", "/", handler)
     client = await aiohttp_client(app)
-    resp = await client.get("/")
-    assert 201 == resp.status
-    txt = await resp.text()
-    assert "OK[MIDDLEWARE]" == txt
+
+    # Call twice to verify cache works
+    for _ in range(2):
+        resp = await client.get("/")
+        assert 201 == resp.status
+        txt = await resp.text()
+        assert "OK[MIDDLEWARE]" == txt
 
 
 async def test_middleware_handles_exception(loop, aiohttp_client) -> None:
@@ -44,10 +48,13 @@ async def test_middleware_handles_exception(loop, aiohttp_client) -> None:
     app.middlewares.append(middleware)
     app.router.add_route("GET", "/", handler)
     client = await aiohttp_client(app)
-    resp = await client.get("/")
-    assert 501 == resp.status
-    txt = await resp.text()
-    assert "Error text[MIDDLEWARE]" == txt
+
+    # Call twice to verify cache works
+    for _ in range(2):
+        resp = await client.get("/")
+        assert 501 == resp.status
+        txt = await resp.text()
+        assert "Error text[MIDDLEWARE]" == txt
 
 
 async def test_middleware_chain(loop, aiohttp_client) -> None:
@@ -418,7 +425,7 @@ async def test_old_style_middleware(loop, aiohttp_client) -> None:
             continue
         assert re.match(
             "^old-style middleware "
-            '"<function test_old_style_middleware.<locals>.'
+            '"<function .*test_old_style_middleware.<locals>.'
             'middleware_factory at 0x[0-9a-fA-F]+>" '
             "deprecated, see #2252$",
             msg,
@@ -460,7 +467,7 @@ async def test_old_style_middleware_class(loop, aiohttp_client) -> None:
             continue
         assert re.match(
             "^old-style middleware "
-            '"<test_web_middleware.test_old_style_middleware_class.'
+            '"<.*test_web_middleware.test_old_style_middleware_class.'
             "<locals>.Middleware object "
             'at 0x[0-9a-fA-F]+>" deprecated, see #2252$',
             msg,
@@ -514,3 +521,27 @@ async def test_new_style_middleware_method(loop, aiohttp_client) -> None:
     assert 201 == resp.status
     txt = await resp.text()
     assert "OK[new style middleware]" == txt
+
+
+async def test_middleware_does_not_leak(aiohttp_client: AiohttpClient) -> None:
+    async def any_handler(request: web.Request) -> NoReturn:
+        assert False
+
+    class Middleware:
+        @web.middleware
+        async def call(
+            self, request: web.Request, handler: Handler
+        ) -> web.StreamResponse:
+            return await handler(request)
+
+    app = web.Application()
+    app.router.add_route("POST", "/any", any_handler)
+    app.middlewares.append(Middleware().call)
+
+    client = await aiohttp_client(app)
+
+    web_app._cached_build_middleware.cache_clear()
+    for _ in range(10):
+        resp = await client.get("/any")
+        assert resp.status == 405
+    assert web_app._cached_build_middleware.cache_info().currsize < 10
